#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>


///////////////////////////////////////////////////////


union o_data {
	float f;
	int n;
	char c;
	char *s;
	void (*callback)();
};
typedef struct stkObject {
	char flags;
	int type;
	struct stkForm *form;
	int x,y,w,h;
	float r,g,b;
	void *free;		// used to free complex objects
	union o_data *v;	// array of callbacks and values
} stkObject;

typedef struct stkForm {
	char flags;
	int x,y,w,h;
	int max_objects;
	int nr_objects;
	struct stkObject **objects;	// array of object pointers	
	struct stkForm *next, *prev;
} stkForm;


///////////////////////////////////////////////////////


// reduce the number of realloc calls for form->objects
#define OBJECTS_PER_BLOCK 5


// object/form visibility flag  
#define STK_FLAGS_VISIBLE (1<<0)
// object/form enable flag
#define STK_FLAGS_ENABLED (1<<1)

// object has user defined color
#define STK_FLAGS_COLOR (1<<2)
// form has a view flag
#define STK_FLAGS_FORM_HAS_VIEW (1<<2)

// button flags
#define STK_FLAGS_BUTTON_PRESSED (1<<7)

// input flags
#define STK_FLAGS_INPUT_RESIZE (1<<7)
#define STK_FLAGS_INPUT_NUMERIC_ONLY (1<<6)


enum object_types {	STK_BUTTON=0, STK_TOGGLE_BUTTON, STK_INPUT,
			STK_VIEW, STK_VIEWGL,
			STK_HSCROLLBAR, STK_VSCROLLBAR, STK_CHECKBOX,
			STK_LABEL, STK_LAST_OBJECT
};


///////////////////////////////////////////////////////


// declared in redraw.c
extern void (*stkRenderForm)( stkForm *form );
extern void (*stkRenderObject[STK_LAST_OBJECT])( stkObject *object, int mouse_over );

// declared in forms.c
extern int stk_nr_forms;
extern stkForm *stk_forms_head;	// array of form pointers

// declared in mouse.c
extern int stk_mouse_x, stk_mouse_y;
extern int stk_mouse_button, stk_mouse_state;
extern stkObject *stk_mouse_over_object;

// declared in init.c
extern int stk_window_ow, stk_window_oh;
extern int stk_window_w, stk_window_h;
extern int stk_window;
extern float stk_window_wr, stk_window_hr;


// declared in input.c
extern stkObject *stk_input_with_focus;
extern int stk_input_cursor, stk_input_left, stk_input_highlight;


///////////////////////////////////////////////////////


// a form is an area that contains objects
		// create an empty form
stkForm		*stkCreateForm	( int x, int y, int w, int h );
		// free form and all objects in it
void		stkFreeForm	( stkForm *form );
		// display form and all object in it, default
void		stkShowForm	( stkForm *form );
		// hide form and all objects in it
void		stkHideForm	( stkForm *form );
		// enable form rendering
void		stkEnableForm	( stkForm *form );
		// disable form rendering, default
void		stkDisableForm	( stkForm *form );


// an object is a widget
		// creates object, only used by object creation functions
stkObject	*stkCreateObject	( int type, stkForm *form, int x, int y, int w, int h, int nr_values );
		// display object, default
void		stkShowObject		( stkObject *object );
		// hide object
void		stkHideObject		( stkObject *object );
		// undo a disable, default
void		stkEnableObject		( stkObject *object );
		// gray out object and ignore
void		stkDisableObject	( stkObject *object );
void		stkSetObjectColor	( stkObject *object, float red, float green, float blue );
void		stkUnsetObjectColor	( stkObject *object );


// button that pushes down and pops up when released
// *press() is called on a mouse_down event
stkObject	*stkCreateButton	( stkForm *form, int x, int y, int w, int h,
					void (*press)(), char *title, float fontsize);


// button that remains down until pressed again
stkObject	*stkCreateToggleButton	( stkForm *form, int x, int y, int w, int h,
					void (*press)(int state), char *up_title,
					char* down_title, int state, float fontsize);
		// return the state of a toggle button
int		stkGetToggleButton	( stkObject *object );
		// set the state of a toggle button
void		stkSetToggleButton	( stkObject *object, int state );


// single-line text input field. *leave_field() is called when the object loses focus
stkObject	*stkCreateInput	( stkForm *form, int x, int y, int w, int h,
				void (*change)(char *text), char *text,
				int max_size, float fontsize );
		// return contents of input field
char		*stkGetInput	( stkObject *object );
		// set contents of input field
void		stkSetInput	( stkObject *object, char *text );
void		stkClearInput	( stkObject *object );


// drawing area, calls *redraw() to update view contents
// calls *key() and *mouse() for all events that take place when view object has focus
// buttons contains a bit for each button, 32bits
// *active_mouse() is called when button is down
// *passive_mouse() is called when buttons are released and mouse moves in view
stkObject	*stkCreateView	( stkForm *form, int x, int y, int w, int h,
				void (*redraw)(),
				void (*key_press)(int state, char ascii),
				void (*mouse_button)(int button, int state, int x, int y),
				void (*mouse_motion)(int x, int y) );
stkObject	*stkCreateViewGL ( stkForm *form, int x, int y, int w, int h,
				void (*redraw)(),
				void (*resize)(int width, int height),
				int *window );


// horizontal and vertical scrollbars. left/right and top/bottom are the min/max values.
// *change() is called when the scrollbar value is changed
stkObject	*stkCreateHScrollbar	( stkForm *form, int x, int y, int w, int h,
					void (*change)(int value), int left, int right, int value );
stkObject	*stkCreateVScrollbar	( stkForm *form, int x, int y, int w, int h,
					void (*change)(int value), int top, int bottom, int value );
		// return value for scrollbar
int		stkGetScrollbar		( stkObject *object );
		// set value for scrollbar
void		stkSetScrollbar		( stkObject *object, int value );


// checked toggle button. *change() is called when clicked
stkObject	*stkCreateCheckbox	( stkForm *form, int x, int y, int w, int h,
					void (*change)(int state), int state );
		// get state for checkbox
int		stkGetCheckbox		( stkObject *object );
		// set state for checkbox
void 		stkSetCheckbox		( stkObject *object, int state );


// display text
stkObject	*stkCreateLabel	( stkForm *form, int x, int y, char *title, float fonsize );


// initialize glut and hook stk calls to glut callbacks
int	stkInitGlut		( int argc, char *argv[], int window_w, int window_h,
				char *window_name, int fullscreen);
void	stkShutdown		( );
void	stkResize		( int width, int height );
void	stkRenderBegin		( );
void	stkRenderEnd		( );

void	stkRedraw		( );
void	stkRedrawForm		( stkForm *form );

stkObject *stkIsOverObject	( int x, int y );
void	stkMouseButton		( int button, int state, int x, int y );
void	stkMouseActive		( int x, int y );
void	stkMousePassive		( int x, int y );
void	stkMouseEntry		( int state );

void	stkKey			( unsigned char key, int x, int y );
void	stkSpecialKey		( int key, int x, int y );
void	stkSpecialKeyUp		( int key, int x, int y );
